from tensorflow.keras.layers import Layer, Dense, Concatenate

from kattention.layers import SelfAttention


class MultiHeadAttention(Layer):

    def __init__(self, heads, **kwargs):
        super(MultiHeadAttention, self).__init__(**kwargs)
        self.heads = heads
        self.attention_heads = [SelfAttention() for _ in range(self.heads)]
        self.concat_heads = Concatenate()
        self.embed = None

    def build(self, input_shape):
        embedding_size = input_shape[-1]
        self.embed = Dense(units=embedding_size, activation='linear')

    def call(self, inputs, **kwargs):
        self_attention_outputs = [attention_head(inputs) for attention_head in self.attention_heads]
        concatenated_heads = self.concat_heads(self_attention_outputs)
        return self.embed(concatenated_heads)



import setuptools

with open('README.md', 'r') as description_file:
    long_description = description_file.read()

setuptools.setup(
    name='kattention',
    version='0.1.0',
    author='Sofiene ALOUINI',
    author_email='sofiene.alouini@gmail.com',
    description='Package implementing different attention mechanisms as tf.keras layers',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/sofienealouini/kattention',
    python_requires='>=3.7',
    install_requires=['tensorflow>=2.2'],
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ]
)



from tensorflow.keras.layers import Layer, Dense, Dot, Softmax
from tensorflow.python.keras.layers import Lambda


class SelfAttention(Layer):

    def __init__(self, **kwargs):
        super(SelfAttention, self).__init__(**kwargs)
        self.dense_q = None
        self.dense_k = None
        self.dense_v = None
        self.dot_context = Dot(axes=(2, 2))
        self.scale = None
        self.normalize = Softmax()
        self.reweight_v = Dot(axes=(2, 1))

    def build(self, input_shape):
        embedding_size = input_shape[-1]
        self.dense_q = Dense(units=embedding_size, activation='linear')
        self.dense_k = Dense(units=embedding_size, activation='linear')
        self.dense_v = Dense(units=embedding_size, activation='linear')
        scaling_factor = 1 / embedding_size
        self.scale = Lambda(lambda x: x * scaling_factor)

    def call(self, inputs, **kwargs):
        queries = self.dense_q(inputs)
        keys = self.dense_k(inputs)
        values = self.dense_v(inputs)
        context_weights = self.dot_context([queries, keys])
        scaled_context_weights = self.scale(context_weights)
        normalized_context_weights = self.normalize(scaled_context_weights)
        return self.reweight_v([normalized_context_weights, values])



from tensorflow.keras.layers import Layer, Add, LayerNormalization, Dense

from kattention.layers import MultiHeadAttention


class Transformer(Layer):

    def __init__(self, attention_heads, **kwargs):
        super(Transformer, self).__init__(**kwargs)
        self.attention_heads = attention_heads
        self.attend = MultiHeadAttention(heads=self.attention_heads)
        self.add_1 = Add()
        self.normalize_1 = LayerNormalization()
        self.dense = None
        self.add_2 = Add()
        self.normalize_2 = LayerNormalization()

    def build(self, input_shape):
        embedding_size = input_shape[-1]
        self.dense = Dense(units=embedding_size, activation='linear')

    def call(self, inputs, **kwargs):
        inputs_with_attention = self.attend(inputs)
        inputs_with_attention_and_residual = self.add_1([inputs_with_attention, inputs])
        normalized_embeddings_with_attention = self.normalize_1(inputs_with_attention_and_residual)
        new_embeddings = self.dense(normalized_embeddings_with_attention)
        new_embeddings_with_residual = self.add_2([normalized_embeddings_with_attention, new_embeddings])
        return self.normalize_2(new_embeddings_with_residual)
